/* ----------------------------------------------------------------------------
**      config.c Configuration file reading routines
** ----------------------------------------------------------------------------
**      TGT                                       TextGUIToolkit Library
**      Copyright (C)   2000-2001       Mateusz Golicz and Tomasz Sterna
**
**      LGPL license
**
** ----------------------------------------------------------------------------
**      MAINTAINER      Mateusz Golicz
**
**      Log:
**      - cleanup
**
*/


#define TGT_PREFS_FILE
#include "tgt.h"
#include <pwd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

int isblank (int c); /* does not appear in my libc includes !! */
int g_error_handling=TGT_ERRORHANDLING_TGT; 
    /* by default, exit() on every fatal error */

/*
 *
 * tgt_set_errorhandling: Change error handling mode
 *
 */
void tgt_set_errorhandling(int t)
{
    g_error_handling=t;
}       	      

/*
 *
 * tgt_conf_hash: Get hash function value from given string
 *
 */
static int tgt_conf_hash(char *str)
{
    int i,l,r;
    for(i=0,l=strlen(str),r=0;i<l;i++) r=(r << 1) + str[i];
    return(r % HASH_SIZE);
}

/* this structures hold internal data of config facility */

/* a single key = value pair: */
struct tgt_value
{
    struct tgt_value *next; /* next in list */
    char *key; /* key string */
    char *string; /* value string */
};

/* section */
struct tgt_section
{
    struct tgt_section *next;  /* next in list */
    char *name; /* section name */
    struct tgt_value * pairs[HASH_SIZE]; /* hashed by key table of all key=value pairs */
};

/* whole file */
struct tgt_prefs  
{
    struct tgt_section * sections[HASH_SIZE];
	/* hashed by name table of all sections in file */
};

/*
 *
 * tgt_get_conf: Get a configuration variable from given config object
 *
 */

char* tgt_get_conf(struct tgt_prefs *prefs, /* Config object */
		   char *section, /* Section name */
		   char *key,     /* Key name */
		   char *def)     /* What to return if not found in file ? */
{
    struct tgt_section * sec; 
    struct tgt_value * pair;
    int k;
    
    if(!prefs) return(def); /* We do not have any config object ! */
    
    k=tgt_conf_hash(section); 
    for(sec=prefs->sections[k];sec!=NULL;sec=sec->next) 
	if(strcmp(section,sec->name)==0) break; /* look for section of given name */

    if(!sec) return(def); /* there is no such section */
    
    k=tgt_conf_hash(key);
    for(pair=sec->pairs[k];pair!=NULL;pair=pair->next) 
        if(strcmp(key,pair->key)==0) break; /* look for key of given name */
	
    if(!pair) return(def); /* there is no such key */
    return(pair->string); /* return value associated with that key */
}

/*
 *
 * tgt_set_conf: Set a configuration variable in given config object
 *
 */

int tgt_set_conf(struct tgt_prefs *prefs,
		 char* section,
		 char* key,
		 char* value)
{
    int k,ret,i;
    struct tgt_section *sec;
    struct tgt_value *pair;
    
    k=tgt_conf_hash(section); ret=0;
    for(sec=prefs->sections[k];sec!=NULL;sec=sec->next)
	if(strcmp(section,sec->name)==0) break;
    /* look for given section */
    
    if(!sec) /* no such section .. */
    {
	ret|=CONF_NOTICE_NOSUCHSECTION; /* notify the caller, that it had to be created */
	sec=(struct tgt_section*) malloc(sizeof(struct tgt_section));
	sec->name=strdup(section); /* copy the section name */
	for(i=0;i<HASH_SIZE;i++) sec->pairs[i]=NULL; /* zero its items hash table */
	sec->next=prefs->sections[k]; /* link into list */
	prefs->sections[k]=sec;
    }
    
    k=tgt_conf_hash(key);
    for(pair=sec->pairs[k];pair!=NULL;pair=pair->next)
	if(strcmp(key,pair->key)==0) break;
    /* look for given key */
    
    if(!pair) /* no such key .. */
    {
	ret|=CONF_NOTICE_NOSUCHKEY; /* notify the caller */
	pair=(struct tgt_value*) malloc(sizeof(struct tgt_value));
	pair->key=strdup(key); /* copy the key name */
	pair->next=sec->pairs[k];
	sec->pairs[k]=pair; /* link it into list */
    }
    else
	free(pair->string); /* free old value */

    pair->string=strdup(value); /* copy new key value */

    return(ret); /* return status */
}

/*
 *
 * tgt_reparse: Change weird characters in string to escape/special sequences
 *
 */
unsigned char *tgt_reparse(unsigned char *string)
{
    unsigned char *ret;
    unsigned char c,d,e;
    int l,i,k;
    
    l=strlen(string);
    ret=(unsigned char*) malloc(l*4+2); /* in the worst case we'll need so much */
    for(i=0,k=0;i<l;i++) /* foreach character in string */
    {
	c=string[i];
	if(!isprint(c)) /* it's nonprintable */
	{
	    d=c >> 4; e=c & 0xf; /* change it to hex digits */
	    if(d<10) d+='0'; else d=d+'a'-10;
	    if(e<10) e+='0'; else e=e+'a'-10; /* and transform to ascii */
	    
	    ret[k++]='\\'; ret[k++]='x'; ret[k++]=d; ret[k++]=e; 
	    /* write the sequence to the final string */
	}
	else
	{
	    if(c=='\\' || c=='"') /* it's a special character */
	    {
		ret[k++]='\\'; ret[k++]=c; /* escape it with \ and write into the final string */
	    }
	    else
		ret[k++]=c; /* leave it alone */
	}
    }
    ret[k++]=0;
    ret=(char*) realloc(ret, k);
    return(ret);
}

/*
 *
 * tgt_parse: Process special sequences in strings (\ and \x..)
 *
 */
unsigned char *tgt_parse(unsigned char *string,int *length,unsigned char terminator)
{
    unsigned char *out;
    unsigned char c,d,e;
    int i,k;

    out=(char*) malloc(strlen(string)); k=0; /* in the worst case we'll need so much */
    for(i=0;string[i]!=0;i++) /* foreach character in string */
    {
	if(string[i]=='\\') /* is it '\' ? */
	{
	    c=string[i+1]; /* get next character .. */
	    if(!c) continue; /* if it's end of string, forget that */
	    if(c=='x') /* x - means hexadecimal code of character */
	    {
	        if((d=string[i+2])) /* get first hex digit */
	        {
	    	    if((e=string[i+3])) /* get second hex digit */
		    {
		        if(d>='a' && d<='f') d=d-'a'+10; else d-='0';
		        if(e>='a' && e<='f') e=e-'a'+10; else e-='0';
		        out[k++]=d << 4 | e; /* calculate character code and write into final string */
			i+=3; /* x<hex_digit><hex_digit> */
			continue; 
		    }
		}
	    }
	    out[k++]=c; i++; /* just an escaped character */
	}
	else
	{
	    if(string[i]==terminator) break; /* we reached end of string */
	    out[k++]=string[i]; /* add that end marker to final string */
	}
    }
    out[k]=0;
    if(length) *length=i;
    return(out);
}

/*
 *
 * tgt_compresstr: Remove spaces from string
 *
 */
char * tgt_compressstr(char *a)
{
    int i,k,l;
    char *ret;
    ret=(char*) malloc(l=strlen(a));
    for(i=0,k=0;i<l;i++) if(!isspace(a[i])) ret[k++]=a[i];
    ret[k]=0; 
    ret=(char*) realloc(ret,k+1);
    return(ret);
}

/*
 *
 * tgt_unload_conf: Free all memory associated with internal config file structures
 *
 */

void tgt_unload_conf(struct tgt_prefs * prefs)
{
    int i,j;
    struct tgt_section *a,*p;
    struct tgt_value *b,*q;
    for(i=0;i<HASH_SIZE;i++)
    {
	for(a=prefs->sections[i];a;a=p)
	{	
	    for(j=0;j<HASH_SIZE;j++)
	    {
		for(b=a->pairs[j];b;b=q)
		{
		    q=b->next;
		    free(b->key);
		    free(b->string);
		    free(b);
		}
	    }
	    p=a->next;
	    free(a->name);
	    free(a);
	}
    }
    free(prefs);
}

/*
 *
 * tgt_save_conf: Save internal config file structures into a real file
 *
 */

int tgt_save_conf(struct tgt_prefs * prefs,char *filename)
{
    int i,j;
    char *str;
    FILE *fh;
    struct tgt_section *a;
    struct tgt_value *b;
    
    if(!(fh=fopen(filename,"w"))) return(0);
    fprintf(fh,"# Auto-generated by config.o/libtgt.so , Written 2000-2001 by Tomasz Sterna & Mateusz Golicz\n");
    /* write copyright */
    
    for(i=0;i<HASH_SIZE;i++)
	for(a=prefs->sections[i];a;a=a->next)
	{	
	    fprintf(fh,"[%s]\n",a->name); /* Section name */
	    for(j=0;j<HASH_SIZE;j++)
		for(b=a->pairs[j];b;b=b->next) /* foreach pair */
		{
		    fprintf(fh,"%s = \"%s\"\n",b->key,str=tgt_reparse(b->string));
		    free(str);
		}
	}

    fclose(fh);
    return(1);
}

/*
 *
 * tgt_load_conf: Load configuration file and store it in internal structures
 *
 */

struct tgt_prefs * tgt_load_conf(char *name)
{
    FILE *fh;
    char *file,*mem,*hdir;
    struct tgt_prefs *ret;
    int i,size,seclen,keylen,l;
    char *section;
    char *value,*key;
    struct passwd *pentry;
    
    ret=(struct tgt_prefs*) malloc(sizeof(struct tgt_prefs)); /* allocate global object */
    for(i=0;i<HASH_SIZE;i++) ret->sections[i]=NULL; /* and clean it's Section Hash Table */
    
    if(!(fh=fopen(name,"r"))) /* try opening the file in cwd */
    {
	if((pentry=getpwuid(getuid()))) /* get passwd entry of current user */
	    if((hdir=pentry->pw_dir)) /* and obtain his homedir path */
	    {
		file=(char*) malloc(strlen(hdir)+strlen(name)+2);
		sprintf(file,"%s/%s",hdir,name); /* compose new file path */
		fh=fopen(file,"r"); /* .. in order to look for that file in ~ */
		free(file);
	    }
	if(!fh) /* still not suceeded ?? */
	{
	    file=(char*) malloc(strlen(name)+8);
	    sprintf(file,"/etc/%s\n",name); /* compose new path */
	    fh=fopen(file,"r"); /* .. and look for that file in /etc */
	    free(file);
	    if(!fh) return(ret); /* file not found - but all is ok, we just do like
	                            it is an empty file*/
	}
    }
    fseek(fh,0,SEEK_END); size=ftell(fh); rewind(fh); /* get file size */
    mem=file=(char*) malloc(size+1); file[size]=0;
    fread(file,size,1,fh);  /* load whole file into memory */
    fclose(fh); /* and close it */

    section=strdup("main"); /* the name of current section */
    for(;;) /* parse loop */
    {
	for(;isblank(*file) || *file=='\n' || *file=='\r';file++); 
	/* skip all white spaces / enters */
	if(!(*file)) break; /* EOF ? */
 	if(*file=='#')
	{
	    if(!(file=index(file,'\n'))) break; /* a comment, skip to EOL */
	    continue;
	}
	if(*file=='[') /* New section */
	{
	    for(seclen=0;file[seclen]!=']' && file[seclen]!=0;seclen++); /* get section length */
	    if(!(file[seclen]))
	    {
		fprintf(stderr,"tgt_load_conf(): Parse error: unable to find matching ']'\n");
		break;
	    }
	    if(section) free(section); /* free old section name */
	    section=(char*) malloc(seclen);
	    snprintf(section,seclen,"%s",file+1); /* compose new one ...*/
	    file+=seclen+1;
	    continue;
	}
	/* it's an assignement then */
        for(keylen=0;file[keylen]!='=' && file[keylen]!=0;keylen++);
	/* get the length of everything before '=' ... */
	if(file[keylen]==0) break; /* eof ? */
	key=file; key[keylen]=0; file+=keylen+1; /* make a key string */
	for(i=0;isblank(file[i]);i++); /* skip all spaces after = */
	if(file[i]=='"') /* quoted string */
	{
	    value=tgt_parse(file+i+1,&l,'"'); /* find matching " and fetch that string */
	    if(file[i+1]==0 || file[i+2]==0) break; 
	    file+=l+i+2;
	}
	else
	{
	    value=tgt_parse(file,&l,'\n'); /* if it's not quoted, then skip to EOL */
	    file+=l;
	}
	    
	key=tgt_compressstr(key); /* remove spaces in string before = */
	tgt_set_conf(ret,section,key,value); /* and pass it as a key to setconf function */
	free(value); free(key); /* free temporary strings */
    }
    free(mem); free(section); /* clean up */
    return(ret);
}

void * g_prefs;

extern int d_seconds,d_micros;

/*
 *
 * tgt_init: Initialize tgt configuration
 *
 */
void tgt_init(void)
{
    g_prefs=(void*) tgt_load_conf(".tgtrc"); /* load rcfile */
    d_seconds=atoi(tgt_getprefs((struct tgt_prefs*) g_prefs,"keyboard","esc_sec","0"));
    d_micros=atoi(tgt_getprefs((struct tgt_prefs*) g_prefs,"keyboard","esc_usec","200000"));
    /* and initialize miscellaneous configuration variables */
}

