/* ----------------------------------------------------------------------------
**      lowlevel.c  functions dealing with the terminal
** ----------------------------------------------------------------------------
**      TGT                                       TextGUIToolkit Library
**      Copyright (C)   2000-2001       Mateusz Golicz and Tomasz Sterna
**
**      LGPL license
**
** ----------------------------------------------------------------------------
**      MAINTAINER      Mateusz Golicz
**
**      Log:
**      - cleanup
**
*/

#include <unistd.h>
#include <stdio.h>
#ifndef TGT_DLTERMCAP
#include <termcap.h>
#else
#include <dlfcn.h>
#endif
#include <memory.h>
#include <string.h>
#include <stdarg.h>
#include "tgt.h"
#include <stdlib.h>
#define TERMSECTIONNAME "terminal"
/* Pointers for functions from termcap (which we load dynamically) */
#ifdef TGT_DLTERMCAP
    char* (*tgetstr)(char *,char*);
    int (*tgetnum)(char *);
    int (*tgetent)(char *,char*);
    char* (*tparam)(const char *,char *,int, ...);
    int (*tputs)(char *,int,int(*)(int));
#endif
#define PRINTF_BUFFER_LENGTH 512

static void *termcap_dlh = NULL;

/* Preferred cursor position after screen update */
int tgt_preferred_x = -1,tgt_preferred_y = -1;

struct tgt_terminal * g_terminal = NULL;

/*
 * tgt_out_param1: Front-end for tparam & tputs for one-argument control strings
 *
 */

static inline void tgt_out_param1(char * fstring,int param)
{
    char tbuffer[40];
    char *buf;
#ifdef TGT_DLTERMCAP
    if(!termcap_dlh)  /* if we do not have termcap */
	printf(fstring,param); /* then just print the string out */
    else
    {
#endif
	buf=tparam(fstring,tbuffer,40,param); /* otherwise format it using tparam */
	tputs(buf,1,putchar); /* and output it to terminal using tputs */
	if(buf!=tbuffer) free(buf); /* in case termcap had to allocate memory for us */
#ifdef TGT_DLTERMCAP
    }
#endif
}


/*
 * tgt_out_param2: Front-end for tparam & tputs for two-argument control strings
 *
 */

static inline void tgt_out_param2(char * fstring,int param1,int param2)
{
    char tbuffer[40];
    char *buf;
#ifdef TGT_DLTERMCAP
    if(!termcap_dlh) 
	printf(fstring,param1+1,param2+1);
    else
    {
#endif
	buf=tparam(fstring,tbuffer,40,param1,param2);
	tputs(buf,1,putchar);
	if(buf!=tbuffer) free(buf);
#ifdef TGT_DLTERMCAP
    }
#endif
}

#ifndef TGT_DLTERMCAP
#define tgt_tputs(a,b,c) tputs(a,b,c)
#else

/*
 * tgt_tputs: Front-end for tputs
 *
 */

inline static void tgt_tputs(char *string,int l,int (*out)(int))
{
    if(termcap_dlh) tputs(string,l,out); else printf(string);
}
#endif


/*
 * tgt_chattr: Change a terminal attribute such as fg & bg colors 
 *
 */

void tgt_chattr(int request, /* Capability , see include/terminal.h */
		int a,int b) /* Arguments */
{

    switch(request)
    {
	case TGT_TA_BGCOLOR: if(g_terminal->bg!=a) { tgt_out_param1(g_terminal->c_bgcolor,a); g_terminal->bg=a; } return;
	case TGT_TA_FGCOLOR: if(g_terminal->fg!=a) { tgt_out_param1(g_terminal->c_fgcolor,a); g_terminal->fg=a; } return;
	case TGT_TA_CLEAR: tgt_tputs(g_terminal->c_clear,1,putchar); return;
	case TGT_TA_BOLD: tgt_tputs(g_terminal->c_active,1,putchar); return;
	case TGT_TA_NORMAL: tgt_tputs(g_terminal->c_inactive,1,putchar); g_terminal->fg=-1; g_terminal->bg=-1; return;
	case TGT_TA_GFX: tgt_tputs(g_terminal->c_graphics,1,putchar); return;
	case TGT_TA_TXT: tgt_tputs(g_terminal->c_text,1,putchar); return;
	case TGT_TA_VISIBLECURSOR: tgt_tputs(g_terminal->c_enablecurs,1,putchar); return;
	case TGT_TA_INVISIBLECURSOR: tgt_tputs(g_terminal->c_disablecurs,1,putchar); return;
	case TGT_TA_CURSOR: tgt_out_param2(g_terminal->c_move,b,a); return;
	case TGT_TA_COLORS: 
	    if(g_terminal->bg!=b) { tgt_out_param1(g_terminal->c_bgcolor,b); g_terminal->bg=(char) b; }
	    if(g_terminal->fg!=a) { tgt_out_param1(g_terminal->c_fgcolor,a); g_terminal->fg=(char) a; }
	    return;
    }
}

/*
 * tgt_int_lowerb: Draw bottom border
 *
 */
void tgt_int_lowerb(tgt_cell * buffer, /* where to start */
		    int l, /* length of our border */
		    tgt_cell attr) /* color, etc */ 
{
    int i;
    tgt_cell p;
    *(buffer++)=TGT_T_FCHAR(attr,TGT_TC_LL); /* |_ character */
    l -= 2; p = TGT_T_FCHAR(attr,TGT_TC_HL); /* horizontal line */
    for(i=0;i<l;i++) *(buffer++) = p;        /* write it l-2 times */
    *(buffer++)=TGT_T_FCHAR(attr,TGT_TC_LR); /* _| character */
    return;
}

/*
 * tgt_int_upperb: Draw top border
 *
 */
void tgt_int_upperb(tgt_cell * buffer,int l,tgt_cell attr)
{
    int i;
    tgt_cell p;
    *(buffer++)=TGT_T_FCHAR(attr,TGT_TC_UL); /* |^ character */
    l -= 2; p = TGT_T_FCHAR(attr,TGT_TC_HL);     /* horizontal line */
    for(i=0;i<l;i++) *(buffer++)=p;          /* write it l-2 times */
    *(buffer++)=TGT_T_FCHAR(attr,TGT_TC_UR); /* ^| character */
    return;
}

/*
 * tgt_int_upperb: Fill buffer with a given cell
 *
 */
void tgt_fill_buffer(tgt_cell * buffer,
		     int xs,int ys, /* buffer size */
		     tgt_cell cell) /* pattern */
{
    register int i,m;
    m = xs*ys; /* calculate buffer size in cells */
    for(i=0;i<m;i++)
	*(buffer++) = cell; /* .. and simply fill it */
}

/*
 * tgt_change_colors: Change attributes part of cells in a buffer
 *
 */
void tgt_change_colors(tgt_cell * buffer,int length,int fg,int bg,int bold)
{
    register int i;
    tgt_cell cell = TGT_T_BOLD(TGT_T_FG(TGT_T_BG(0,bg),fg),bold);

    for(i=0;i<length;i++, buffer++)
	*buffer=((*buffer) & 0xff) | cell; 
    /* leave only character part in LSB, and put new attributes in MSB */
}

/*
 * tgt_blit: Blit a buffer to another one
 *
 */
void tgt_blit(tgt_cell * dest, /* destination buffer */
	      int dxs,         /* and its line size in cells */
	      tgt_cell *src,   /* source buffer */
	      int xs,int ys,   /* and its dimensions */
	      int xp,int yp)   /* where to blit the smaller buffer */
{
    register int dxsd, i;
    dest += yp*dxs+xp; /* offset from the beginning of buffer */
    dxsd=xs << 1;      /* line size of source buffer in bytes */
    for(i=0;i<ys;i++)
    {
	memcpy(dest,src,dxsd); /* copy the line */
	dest+=dxs; /* go to the next one */
	src+=xs;   /* go to the next one */
    }
}

static unsigned char tgt_printf_buffer[PRINTF_BUFFER_LENGTH+1];

/*
 * tgt_printf: Perform a nprintf() into a cell buffer
 *
 */
int tgt_printf(tgt_cell *buffer, /* where to blit the string */
	       int max, /* maximal length of our string */
	       tgt_cell attr, /* attributes' part of cells that the function'll modify */
	       char *format,...) /* format string and its arguments */
{
    int n, i;
    va_list vl;
    va_start(vl,format);
    n=vsnprintf(tgt_printf_buffer,max+1,format,vl); /* printf the string into temporary buffer */
    for(i=0;i<n;i++) *(buffer++)=TGT_T_FCHAR(attr,tgt_printf_buffer[i]);
    /* ... and blit it into our cell buffer */
    va_end(vl);
    return(n);
}

/*
 * tgt_flprintf: Perform a nprintf() into a cell buffer and fill remaining
 *               space with ' '
 *
 */
void tgt_flprintf(tgt_cell *buffer,int max,tgt_cell attr,char *format,...)
{
    int n,i;
    va_list vl;
    va_start(vl,format);
    n=vsnprintf(tgt_printf_buffer,max+1,format,vl);
    for(i=0;i<n;i++) *(buffer++)=TGT_T_FCHAR(attr,tgt_printf_buffer[i]);
    n=max-n;
    if(n>0) for(i=0;i<n;i++) *(buffer++)=TGT_T_FCHAR(attr,' ');
    va_end(vl);
}

/*
 * tgt_update_screen: Synchronize terminal contents with TGT screen buffers
 *
 */
void tgt_update_screen(void)
{
    tgt_cell * buffer = g_terminal->terminal_buffer; /* current real contents */
    tgt_cell * screen = g_terminal->contents_buffer; /* current terminal contents */
    int y,ys=g_terminal->y_size,linelength=g_terminal->x_size << 1,xs=g_terminal->x_size;
    int x,xl,inc;
    int bold,fg,bg;
    int c_gfx=-1,c_bold=-1,c_fg=-1,c_bg=-1;
    tgt_cell cell;
    int c;    
    
    for(y=0;y<ys;y++) /* foreach line */
    {
	if(memcmp(buffer,screen,linelength)) /* did it change ? */
	{
	    for(x=0;;x++) if(buffer[x]!=screen[x]) break; 
					    /* First different character */
	    for(xl=xs-1;;xl--) if(buffer[xl]!=screen[xl]) break;
					    /* Last different character */
	    xl -= x; /* how many chars to copy */
	    buffer+=x; screen+=x; /* seek to the first different char */
	    tgt_chattr(TGT_TA_CURSOR,x,y); /* set cursor position */
//	    printf("@%d,%d ", x,y);
	    inc = xs - x - xl - 1;/* how many chars will remain to end-of-line ? */
	    for(x=0;x<=xl;x++) /* foreach character in line area to copy */
	    {
		cell=*(buffer++); /* get the cell */
		*(screen++)=cell; /* and update the terminal contents buffer */

		/* and then write it to terminal */
		if((bold=TGT_T_GETBOLD(cell)) != c_bold) /* is the bold state different from previous ? */
		{
		    if((c_bold=bold))
			tgt_chattr(TGT_TA_BOLD,0,0);
		    else
		    {
			tgt_chattr(TGT_TA_NORMAL,0,0);
			c_fg=-1; c_bg=-1; /* TA_NORMAL is usually \033[0m, which resets all attributes */
		    }
		}

		/* are the colors just like in previous character ? */
		if((fg=TGT_T_GETFG(cell)) != c_fg) tgt_chattr(TGT_TA_FGCOLOR,c_fg=fg,0);
		if((bg=TGT_T_GETBG(cell)) != c_bg) tgt_chattr(TGT_TA_BGCOLOR,c_bg=bg,0);

		/* get the character */
		c=TGT_T_GETCHAR(cell);
		
		if(c<32) /* it's non-printable */
		{
		    if(c<8) /* it's a graphic char (see include/terminal.h) */
		    {
			if(c_gfx!=1) /* is the graphics on ? */
			{
			    c_gfx=1;
			    tgt_chattr(TGT_TA_GFX,0,0);
			}
			putchar(g_terminal->gfx_set[c]);
		    }
		    else
			putchar(' '); /* illegal char, output space */
		}
		else
		{
		    if(c_gfx) /* is the graphics off ? */
		    {
			tgt_chattr(TGT_TA_TXT,0,0);
			c_gfx=0;
		    }
		    putchar(c);
		}
	    }
	    buffer+=inc; /* go to the end of line */
	    screen+=inc;
	}
	else
	{
	    buffer+=g_terminal->x_size; /* the line has not changed */
	    screen+=g_terminal->x_size; /* so go to the end of it */
	}
    }
    if(tgt_preferred_x!=-1) /* one of tgt routines wanted to set the cursor position ? */
    {
	/* so do as it wished */
	tgt_chattr(TGT_TA_CURSOR,tgt_preferred_x,tgt_preferred_y);
	tgt_preferred_x=-1;
    }
    fflush(stdout); /* flush libc buffers */
}

/*
 * tgt_destroyterminal: Free memory allocated while creating a terminal
 *                      description
 *
 */
void tgt_destroyterminal(struct tgt_terminal *tterm)
{
    /* Free control strings */
    free(tterm->c_graphics); free(tterm->c_text); free(tterm->c_clear);
    free(tterm->c_fgcolor); free(tterm->c_bgcolor); free(tterm->c_active);
    free(tterm->c_inactive); free(tterm->c_move); 
    /* Destroy multiple key sequences lookup tables */
    tgt_destroylookuptable(tterm->lookup_root);
    /* free the structure memory itself */
    free(tterm);
}

/*
 * tgt_changedefcolors: Change default terminal colors
 *
 */
void tgt_changedefcolors(struct tgt_terminal * term,unsigned char fg,unsigned char bg)
{
    term->color_fg=fg;
    term->color_bg=bg;
}

/*
 * tgt_term_addkey: Add a multiple key sequence into a terminal description
 *                  The sequence is passed directly or fetched from termcap
 *
 */
void tgt_term_addkey(struct tgt_terminal * term,
		     char *capability, /* termcap capability describing that sequence*/
		     char *def, /* what to set if it's not found in termcap or termcap not present */
		     int key) /* what to return to upper-level functions when this sequence is detected in input stream */
{
    char *str = NULL;
    char kbbuffer[TGT_MAX_SEQ];
    
    if(!term) term = g_terminal;
#ifdef TGT_DLTERMCAP
    if(termcap_dlh)
#endif
	str=tgetstr(capability,NULL); /* try reading the sequence from termcap */

    if(str==NULL) /* did it work ? */
    	snprintf(kbbuffer,TGT_MAX_SEQ-1,def); /* discard control sequences from the default string and save it under kbbuffer */
    else
    {
	snprintf(kbbuffer,TGT_MAX_SEQ-1,str); /* discard control sequences from the string termcap returned and save it under kbbuffer */
	free(str);
    }
    tgt_addkeyseq(term->lookup_root, kbbuffer, key); /* see lowlevel/keytree.c and doc/en/LOOKUP-TABLES */
}
    extern void * g_prefs;

/*
 * tgt_term_addkey_int: Add a multiple key sequence into a terminal description
 *			The sequence is passed directly, fetched from termcap or
 *			taken from tgt rcfile.
 *
 */
void tgt_term_addkey_int(int te, /* is termcap present ? */
			 struct tgt_terminal *term,
			 char *capability, /* termcap name */
			 char *def, /* default */
			 char *conf_name,  /* tgt rcfile name */
			 int key) /* what to return if the sequence is detected */
{
    char *str;
    char *pstr;
    char kbbuffer[TGT_MAX_SEQ];
    str = NULL;

    /* try fetching from tgtrc */
    if((pstr=tgt_getprefs(g_prefs,TERMSECTIONNAME,conf_name,NULL)))
	def=pstr;
    else
	if(te) str=tgetstr(capability,NULL); /* not found? then check in termcap */

    if(!str)
    	snprintf(kbbuffer,TGT_MAX_SEQ-1,def); /* nothing found? use default */
    else
    {
	snprintf(kbbuffer,TGT_MAX_SEQ-1,str);
	free(str);
    }
    tgt_addkeyseq(term->lookup_root,kbbuffer,key); /* see lowlevel/keytree.c and doc/en/LOOKUP-TABLES */
}

/*
 * tgt_tgetstrd: tgetstr() replacement: get a capability from termcap, or
 *               use tgt rcfile setting or use default.
 *
 */
char *tgt_tgetstrd(int te, /* is termcap present ? */
		   char *name, /* termcap name */
		   char *def,  /* default */
		   char *conf_name) /* rcfile name */
{
    char *ret;

    if((ret=tgt_getprefs(g_prefs,TERMSECTIONNAME,conf_name,NULL))) 
	return(strdup(ret));
    if(te && !ret) ret=tgetstr(name,NULL);
    if(!ret) ret=strdup(def);
    return(ret);
}


/*
 * tgt_setscreen: Allocate and configure a terminal description
 *
 */
struct tgt_terminal * tgt_setscreen(char *name) 
{
#ifdef TGT_DLTERMCAP
    int te=0; /* is termcap available ? */
#else
    int te=1; /* in case of statically linked termcap it's always available, i guess */
#endif
    int n;
    char *buffer, *sb;
    struct tgt_terminal *ret;
#ifdef TGT_DLTERMCAP
    void *dlh = NULL;
    
    if(!termcap_dlh) /* did we open the termcap ? */
    {
	dlh=dlopen("libtermcap.so",RTLD_LAZY);
	if(!dlh) dlh=dlopen("libtermcap.so.2",RTLD_LAZY);
	if(dlh)
	{
	    /* get functions' pointers */
	    tgetent=(int(*)(char *,char*)) dlsym(dlh,"tgetent");
	    tgetnum=(int(*)(char *)) dlsym(dlh,"tgetnum");
	    tgetstr=(char*(*)(char *,char*)) dlsym(dlh,"tgetstr");
	    tparam=(char*(*)(const char *,char*,int,...)) dlsym(dlh,"tparam");
	    tputs=(int(*)(char *,int,int(*)(int))) dlsym(dlh,"tputs");
	    /* one of functions was not found, weird thing ... */
	    if(!tgetent || !tgetnum || !tgetstr || !tputs)
	    {
		dlclose(dlh);
		dlh=NULL;
	    }
	    /* everything seems ok, but there is no tparam,
	       which means ncurses wants to trick us */
	    if(!tparam)
	    {
		dlclose(dlh);
		dlh=NULL;
		printf("You DO NOT HAVE a REAL termcap, your libtermcap.so is just a ncurses termcap replacement!\n");
	    }
	}
    }
#endif


/* Read $TERM if neccessary */
    if(name==NULL) name=(char*) getenv("TERM");
    if(name==NULL) name="linux";

    buffer=(char*) malloc(2048);  /* required by POSIX */
#ifdef TGT_DLTERMCAP
    if(dlh) if(name) te=tgetent(buffer,name); /* open our terminal description */
#else
    if(name) te=tgetent(buffer,name);
#endif

    if(te!=1) /* terminal description not available, or simply no termcap */
    {
	free(buffer);
	printf("Termcap library or termcap entry NOT FOUND\n");
	printf("Attempting to use config file directives or defaults\n");
	printf("Use ignore_termcap_type=yes in terminal section in .tgtrc if this message annoys you ;)\n");
	if(!tgt_getprefs(g_prefs,TERMSECTIONNAME,"ignore_termcap_type",NULL))
	{
	    printf("Press ENTER to continue");
	    getchar();
	}
	te=0;
    }

    ret=(struct tgt_terminal *) malloc(sizeof(struct tgt_terminal));

/* defaults: gray text and blue background */
    ret->color_bg=4; ret->color_fg=7;

/* terminal size */

    if((sb=getenv("COLUMNS")))
        ret->x_size=atoi(sb);
    else
        if(te) ret->x_size=tgetnum("co"); else ret->x_size=80;

    if((sb=getenv("LINES")))
        ret->y_size=atoi(sb);
    else
        if(te) ret->y_size=tgetnum("li"); else ret->y_size=25;


    if((n=atoi(tgt_getprefs(g_prefs,TERMSECTIONNAME,"columns","0")))) ret->x_size=n;
    if((n=atoi(tgt_getprefs(g_prefs,TERMSECTIONNAME,"lines","0")))) ret->y_size=n;
    
/* get terminal capabilities .. */
    ret->c_clear=tgt_tgetstrd(te,"cl","\033[H\033[J","clear");
    ret->c_move=tgt_tgetstrd(te,"cm","\033[%d;%dH","move");
    ret->c_graphics=tgt_tgetstrd(te,"as","\x0e","gfx_start");
    ret->c_text=tgt_tgetstrd(te,"ae","\x0f","gfx_end");
    ret->c_fgcolor=tgt_tgetstrd(te,"AF","\033[3%dm","ch_fg");
    ret->c_bgcolor=tgt_tgetstrd(te,"AB","\033[4%dm","ch_bg");
    ret->c_active=tgt_tgetstrd(te,"md","\033[1m","bold");
    ret->c_inactive=tgt_tgetstrd(te,"me","\033[0m","normal");
    ret->c_enablecurs=tgt_tgetstrd(te,"ve","\033[?25h","cursenable");
    ret->c_disablecurs=tgt_tgetstrd(te,"vi","\033[?25l","cursdisable");

/* get multiple-key sequences (such as \033[A) */   
    ret->lookup_root=tgt_initroottable();
    tgt_term_addkey_int(te,ret,"ku","\033[A","c_up",TGT_KEY_UP);
    tgt_term_addkey_int(te,ret,"kd","\033[B","c_down",TGT_KEY_DOWN);
    tgt_term_addkey_int(te,ret,"kl","\033[D","c_left",TGT_KEY_LEFT);
    tgt_term_addkey_int(te,ret,"kr","\033[C","c_right",TGT_KEY_RIGHT);
    tgt_term_addkey_int(te,ret,"kD","\x1b\x5b\x33\x7e","delete",TGT_KEY_DELETE);
    tgt_term_addkey_int(te,ret,"kb","\x7f","backspace",TGT_KEY_BKSPC);
    tgt_term_addkey_int(te,ret,"kI","\x1b\x5b\x32\x7e","insert",TGT_KEY_INSERT);
    tgt_term_addkey_int(te,ret,"kh","\x1b\x5b\x31\x7e","home",TGT_KEY_HOME);
    tgt_term_addkey_int(te,ret,"@7","\x1b\x5b\x34\x7e","end",TGT_KEY_END);
    tgt_term_addkey_int(te,ret,"kP","\x1b\x5b\x35\x7e","pgup",TGT_KEY_PGUP);
    tgt_term_addkey_int(te,ret,"kN","\x1b\x5b\x36\x7e","pgdn",TGT_KEY_PGDN);
    tgt_term_addkey_int(te,ret,"WM","\x17","windowmove",TGT_KEY_WINDOWMOVE);

/* get graphics characters */
    memcpy(ret->gfx_set,tgt_getprefs(g_prefs,TERMSECTIONNAME,"gfx","qxutlkmj"),8);

/* current fg/bg, AFAIR not used in tgt at the moment */
    ret->fg=-1; ret->bg=-1;

/* allocate current and real screen contents buffers */
    ret->terminal_buffer=(tgt_cell*) malloc(sizeof(tgt_cell)*ret->x_size*ret->y_size);
    ret->contents_buffer=(tgt_cell*) malloc(sizeof(tgt_cell)*ret->x_size*ret->y_size);

/* ... and clear it */
    tgt_fill_buffer(ret->terminal_buffer,ret->x_size,ret->y_size,TGT_T_BUILDCELL(ret->color_fg,ret->color_bg,0,0,' '));
    tgt_fill_buffer(ret->contents_buffer,ret->x_size,ret->y_size,0);

#ifdef TGT_DLTERMCAP
    termcap_dlh=dlh;
#endif
    g_terminal = ret;
    return(ret);
}


/*
 * tgt_fatal_error: Clear screen, print an error message and perform exit()
 *
 */
void tgt_fatal_error(char *string)
{
    if(g_terminal)
    {
	tgt_chattr(TGT_TA_TXT,0,0);
	tgt_chattr(TGT_TA_NORMAL,0,0);
	tgt_chattr(TGT_TA_CLEAR,0,0);
    }
    else
	printf("\033[0m\x0f\033[H\033[J");
	
    fflush(stdout);
    fprintf(stderr,"tgt: Fatal error: %s\n",string);
    fflush(stderr);
    tgt_rawcon(0);
    exit(0);
}

struct tgt_terminal * tgt_getterminal(void)
{
    return(g_terminal);
}
